Neko 1.99.7
A portable framework for high-order spectral element flow simulations
Loading...
Searching...
No Matches
opr_conv1.hip
Go to the documentation of this file.
1/*
2 Copyright (c) 2021-2026, The Neko Authors
3 All rights reserved.
4
5 Redistribution and use in source and binary forms, with or without
6 modification, are permitted provided that the following conditions
7 are met:
8
9 * Redistributions of source code must retain the above copyright
10 notice, this list of conditions and the following disclaimer.
11
12 * Redistributions in binary form must reproduce the above
13 copyright notice, this list of conditions and the following
14 disclaimer in the documentation and/or other materials provided
15 with the distribution.
16
17 * Neither the name of the authors nor the names of its
18 contributors may be used to endorse or promote products derived
19 from this software without specific prior written permission.
20
21 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
22 "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
23 LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
24 FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
25 COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
26 INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
27 BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
28 LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
29 CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
30 LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
31 ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
32 POSSIBILITY OF SUCH DAMAGE.
33*/
34
35#include <string.h>
36#include <stdlib.h>
37#include <stdio.h>
38#include <hip/hip_runtime.h>
40#include <device/hip/check.h>
41#include "conv1_kernel.h"
42#include "elem_block_tune.h"
43
44extern "C" {
45 #include <common/neko_log.h>
46}
47
48template < const int >
49int tune_conv1(void *du, void *u,
50 void *vx, void *vy, void *vz,
51 void *dx, void *dy, void *dz,
52 void *drdx, void *dsdx, void *dtdx,
53 void *drdy, void *dsdy, void *dtdy,
54 void *drdz, void *dsdz, void *dtdz,
55 void *jacinv, int *nel, int *gdim, int *lx, int *eb_sel,
56 int *ch_sel);
57
58extern "C" {
59
63 void hip_conv1(void *du, void *u,
64 void *vx, void *vy, void *vz,
65 void *dx, void *dy, void *dz,
66 void *drdx, void *dsdx, void *dtdx,
67 void *drdy, void *dsdy, void *dtdy,
68 void *drdz, void *dsdz, void *dtdz,
69 void *jacinv, int *nel, int *gdim, int *lx) {
70
71 static int autotune[17] = { 0 };
72 /* elements per block candidate chosen by the tuner */
73 static int autotune_eb[17] = { 0 };
74 /* chunk candidate chosen for the 1d variant */
75 static int autotune_ch[17] = { 0 };
76
77 const dim3 nthrds_1d(1024, 1, 1);
78 const dim3 nthrds_kstep((*lx), (*lx), 1);
79 const dim3 nblcks((*nel), 1, 1);
80
81#define CASE_1D(LX, C) \
82 hipLaunchKernelGGL( HIP_KERNEL_NAME( \
83 conv1_kernel_1d<real, LX, NEKO_CHUNKS(LX, C)> ), \
84 nblcks, NEKO_CHUNKS_NTHRDS(LX, C), 0, \
85 (hipStream_t) glb_cmd_queue, \
86 (real *) du, (real *) u, \
87 (real *) vx, (real *) vy, (real *) vz, \
88 (real *) dx, (real *) dy, (real *) dz, \
89 (real *) drdx, (real *) dsdx, (real *) dtdx, \
90 (real *) drdy, (real *) dsdy, (real *) dtdy, \
91 (real *) drdz, (real *) dsdz, (real *) dtdz, \
92 (real *) jacinv); \
93 HIP_CHECK(hipGetLastError());
94
95/* Runtime dispatch onto the tuned chunk candidate */
96#define CASE_1D_SEL(LX, SEL) \
97 switch (SEL) { \
98 case 0: CASE_1D(LX, 0); break; \
99 case 1: CASE_1D(LX, 1); break; \
100 case 2: CASE_1D(LX, 2); break; \
101 default: CASE_1D(LX, 3); break; \
102 }
103
104#define CASE_KSTEP(LX, C) \
105 hipLaunchKernelGGL( HIP_KERNEL_NAME( \
106 conv1_kernel_kstep<real, LX, NEKO_EB(LX, C)> ), \
107 NEKO_EB_NBLCKS(*nel, LX, C), NEKO_EB_NTHRDS(LX, C), 0, \
108 (hipStream_t) glb_cmd_queue, \
109 (real *) du, (real *) u, \
110 (real *) vx, (real *) vy, (real *) vz, \
111 (real *) dx, (real *) dy, (real *) dz, \
112 (real *) drdx, (real *) dsdx, (real *) dtdx, \
113 (real *) drdy, (real *) dsdy, (real *) dtdy, \
114 (real *) drdz, (real *) dsdz, (real *) dtdz, \
115 (real *) jacinv, *nel); \
116 HIP_CHECK(hipGetLastError());
117
118/* Runtime dispatch onto the tuned candidate */
119#define CASE_KSTEP_SEL(LX, SEL) \
120 switch (SEL) { \
121 case 0: CASE_KSTEP(LX, 0); break; \
122 case 1: CASE_KSTEP(LX, 1); break; \
123 default: CASE_KSTEP(LX, 2); break; \
124 }
125
126#define CASE(LX) \
127 case LX: \
128 if(autotune[LX] == 0 ) { \
129 autotune[LX]=tune_conv1<LX>(du, u, \
130 vx, vy, vz, \
131 dx, dy, dz, \
132 drdx, dsdx, dtdx, \
133 drdy, dsdy, dtdy, \
134 drdz, dsdz, dtdz, \
135 jacinv, nel, gdim, lx, &autotune_eb[LX], \
136 &autotune_ch[LX]); \
137 } else if (autotune[LX] == 1 ) { \
138 CASE_1D_SEL(LX, autotune_ch[LX]); \
139 } else if (autotune[LX] == 2 ) { \
140 CASE_KSTEP_SEL(LX, autotune_eb[LX]); \
141 } \
142 break
143
144#define CASE_LARGE(LX) \
145 case LX: \
146 CASE_KSTEP(LX, 0); \
147 break
148
149
150 if ((*lx) < 11) {
151 switch(*lx) {
152 CASE(2);
153 CASE(3);
154 CASE(4);
155 CASE(5);
156 CASE(6);
157 CASE(7);
158 CASE(8);
159 CASE(9);
160 CASE(10);
161 default:
162 {
163 fprintf(stderr, __FILE__ ": size not supported: %d\n", *lx);
164 exit(1);
165 }
166 }
167 }
168 else {
169 switch(*lx) {
170 CASE_LARGE(11);
171 CASE_LARGE(12);
172 CASE_LARGE(13);
173 CASE_LARGE(14);
174 CASE_LARGE(15);
175 CASE_LARGE(16);
176 default:
177 {
178 fprintf(stderr, __FILE__ ": size not supported: %d\n", *lx);
179 exit(1);
180 }
181 }
182 }
183 }
184}
185
186template < const int LX >
187int tune_conv1(void *du, void *u,
188 void *vx, void *vy, void *vz,
189 void *dx, void *dy, void *dz,
190 void *drdx, void *dsdx, void *dtdx,
191 void *drdy, void *dsdy, void *dtdy,
192 void *drdz, void *dsdz, void *dtdz,
193 void *jacinv, int *nel, int *gdim, int *lx, int *eb_sel, int *ch_sel) {
196 int best1 = 0;
198 const int rounds = neko_tune_rounds();
199 const int iters = neko_tune_iters();
200 const int sweep = neko_eb_sweep();
201 int best = 0;
202 int retval;
203
204 for (int c = 0; c < NEKO_EB_CANDIDATES; c++) {
206 }
207 for (int c = 0; c < NEKO_CHUNKS_CANDIDATES; c++) {
209 }
210
211 const dim3 nthrds_1d(1024, 1, 1);
212 const dim3 nthrds_kstep((*lx), (*lx), 1);
213 const dim3 nblcks((*nel), 1, 1);
214 const hipStream_t stream = (hipStream_t) glb_cmd_queue;
215
216 char *env_value = NULL;
217 char neko_log_buf[80];
218
219 env_value=getenv("NEKO_AUTOTUNE");
220
221 sprintf(neko_log_buf, "Autotune conv1 (lx: %d)", *lx);
223
224 *eb_sel = 0;
225 *ch_sel = 0;
226
227 if(env_value) {
228 if( !strcmp(env_value,"1D") ) {
231 sprintf(neko_log_buf,"Set by env : 1 (1D, %d chunk)",
235 return 1;
236 } else if( !strcmp(env_value,"KSTEP") ) {
237 *eb_sel = neko_eb_env();
239 sprintf(neko_log_buf,"Set by env : 2 (KSTEP, %d elem/block)",
243 return 2;
244 } else {
245 sprintf(neko_log_buf, "Invalid value set for NEKO_AUTOTUNE");
247 }
248 }
249
252 /* Warm every variant before timing anything: each specialisation has to be
253 resident and the clocks at steady state, or whichever is timed first is
254 measured on a colder part */
255 for (int i = 0; i < NEKO_TUNE_WARMUP; i++) {
256 CASE_1D(LX, 0);
257 CASE_1D(LX, 1);
258 CASE_1D(LX, 2);
259 CASE_1D(LX, 3);
260 CASE_KSTEP(LX, 0);
261 if (sweep) {
262 CASE_KSTEP(LX, 1);
263 CASE_KSTEP(LX, 2);
264 }
265 }
266
267 /* Interleaved rounds, best time per variant */
268 for (int r = 0; r < rounds; r++) {
274 if (sweep) {
277 }
278 }
279
283 *eb_sel = best;
284 *ch_sel = best1;
285
286 if (time1[best1] < time2[best]) {
287 retval = 1;
288 } else {
289 retval = 2;
290 }
291
292 /* Leave the chosen kernel's output in place: the tuner stands in for a real
293 evaluation and the variants do not sum in the same order */
294 if (retval == 1) {
296 } else {
298 }
299
300 if (retval == 1) {
301 sprintf(neko_log_buf, "Chose : 1 (1D, %d chunk)",
303 } else {
304 sprintf(neko_log_buf, "Chose : 2 (KSTEP, %d elem/block)",
306 }
309 return retval;
310}
__global__ void ale_add_kinematics_kernel(const int n, T *__restrict__ wx, T *__restrict__ wy, T *__restrict__ wz, const T *__restrict__ x_ref, const T *__restrict__ y_ref, const T *__restrict__ z_ref, const T *__restrict__ phi, const T *__restrict__ x, const T *__restrict__ y, const T *__restrict__ z, const kinematics_params_t kin_params)
__global__ void T *__restrict__ T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ drdy
__global__ void T *__restrict__ T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ drdz
__global__ void T *__restrict__ T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ dsdz
__global__ void T *__restrict__ T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ dsdy
const int i
__global__ void T *__restrict__ T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ dtdy
__global__ void T *__restrict__ T *__restrict__ const T *__restrict__ u
__global__ void T *__restrict__ T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ dx
__global__ void T *__restrict__ T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ drdx
__global__ void T *__restrict__ T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ dtdz
__global__ void T *__restrict__ T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ dsdx
__global__ void T *__restrict__ T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ dtdx
__global__ void T *__restrict__ T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ dz
__global__ void T *__restrict__ T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ dy
__global__ void T *__restrict__ T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ jacinv
__global__ void const T *__restrict__ const T *__restrict__ const T *__restrict__ const T *__restrict__ vz
__global__ void const T *__restrict__ const T *__restrict__ vx
__global__ void const T *__restrict__ const T *__restrict__ const T *__restrict__ vy
#define NEKO_CHUNKS_CANDIDATES
Definition elem_block.h:125
#define NEKO_EB_CANDIDATES
Definition elem_block.h:63
#define NEKO_EB_SEL(LX, SEL)
Definition elem_block.h:108
#define NEKO_CHUNKS_SEL(LX, SEL)
Definition elem_block.h:147
#define NEKO_TUNE_TIME(T, LAUNCH, LX, C, ITERS)
static int neko_eb_env()
static int neko_tune_rounds()
#define NEKO_TUNE_LOG(LX, T1, T2)
#define NEKO_TUNE_INIT
#define NEKO_TUNE_BEST(T, BEST, N)
static int neko_tune_iters()
static int neko_chunks_env()
static int neko_eb_sweep()
#define NEKO_TUNE_WARMUP
#define HIP_CHECK(err)
Definition check.h:8
void log_error(char *msg)
void log_message(char *msg)
void log_end_section()
void log_section(char *msg)
int tune_conv1(void *du, void *u, void *vx, void *vy, void *vz, void *dx, void *dy, void *dz, void *drdx, void *dsdx, void *dtdx, void *drdy, void *dsdy, void *dtdy, void *drdz, void *dsdz, void *dtdz, void *jacinv, int *nel, int *gdim, int *lx, int *eb_sel, int *ch_sel)
#define CASE_KSTEP_SEL(LX, SEL)
#define CASE(LX)
#define CASE_1D_SEL(LX, SEL)
void hip_conv1(void *du, void *u, void *vx, void *vy, void *vz, void *dx, void *dy, void *dz, void *drdx, void *dsdx, void *dtdx, void *drdy, void *dsdy, void *dtdy, void *drdz, void *dsdz, void *dtdz, void *jacinv, int *nel, int *gdim, int *lx)
Definition opr_conv1.hip:63
#define CASE_KSTEP(LX, C)
#define CASE_LARGE(LX)
#define CASE_1D(LX, C)